#include <cstdio>       /* stderr          */
#include <cstdlib>      /* exit            */
#include <sys/socket.h> /* accept          */
#include <sys/types.h>  /* fd_set          */
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <vector>

#include "soundtypes.h"
#include "serverfun.h"
#include "mixerstream.hh"
#include "logfun.h"
#include "mixer.h"
#include "argh.hh"

class Mixer
{
    vector<MixerInput> inputs;
    vector<MixerOutput> outputs;
    vector<Waiter> waiters;
    int listfd;
    int humpid;
    int humfd;
public:
    Mixer(int port)
    {
    	int pip[2];
    	pipe(pip);
        humpid = fork();
        if(!humpid)
        {
        	dup2(pip[1], 1);
        	close(pip[0]); close(pip[1]);
        	execlp("mixeridle", "mixeridle", NULL);
        }
        close(pip[1]);
        humfd = pip[0];
        
        listfd = MakeListenerSocket(port);
    }
    
    void tick();
};

void Mixer::tick()
{
    fd_set ifds; FD_ZERO(&ifds);
    fd_set ofds; FD_ZERO(&ofds);
    int topfd = listfd;
    FD_SET(listfd, &ifds);
    
    bool panic = false;
    for(unsigned a=0; a<outputs.size(); ++a)
        if(outputs[a].panic())
        {
            /* panic tarkoittaa, ettei outputtiin mahdu lis tavaraa.
             * tllin pit pit inputit holdissa ja odotella kunnes
             * outputti on synyt sen mit sille on sytetty.
             */
            panic = true;
        }
    if(!panic)
        for(unsigned a=0; a<inputs.size(); ++a)
        {
            int fd = inputs[a].getfd();
            if(fd > topfd)topfd = fd;
            FD_SET(fd, &ifds);
        }
    
    for(unsigned a=0; a<outputs.size(); ++a)
    {
        int fd = outputs[a].getfd();
        if(fd > topfd)topfd = fd;
        if(outputs[a].pending() || !inputs.size())
            FD_SET(fd, &ofds);
    }
    for(unsigned a=0; a<waiters.size(); ++a)
    {
        int fd = waiters[a].getfd();
        if(fd > topfd)topfd = fd;
        FD_SET(fd, &ifds);
    }
    
    select(topfd+1, &ifds, &ofds, NULL, NULL);
    
    if(FD_ISSET(listfd, &ifds))
    {
        struct sockaddr_in They;
        unsigned AddrLen = sizeof They;
        
        int p = accept(listfd, (struct sockaddr *)&They, &AddrLen);
        
        //DumpTime();
        //fprintf(stderr, "Got connection (fd %d)\n", p);
        
        waiters.push_back(Waiter(p, They, AddrLen));
    }
    
    for(unsigned a=0; a<inputs.size(); ++a)
        if(FD_ISSET(inputs[a].getfd(), &ifds))
            inputs[a].Evaluate();

    for(unsigned a=0; a<waiters.size(); ++a)
        if(FD_ISSET(waiters[a].getfd(), &ifds))
            waiters[a].Evaluate();

    for(unsigned a=0; a<outputs.size(); ++a)
        if(FD_ISSET(outputs[a].getfd(), &ofds))
            outputs[a].Evaluate();
    
    vector<Waiter>::iterator wi;
    for(wi = waiters.begin(); wi != waiters.end(); )
        if(wi->isinput()) { inputs.push_back(*wi); waiters.erase(wi); }
        else if(wi->isoutput()) { outputs.push_back(*wi); waiters.erase(wi); }
        else ++wi;
        
    bool problems = false;

    vector<MixerInput>::iterator ii;
    for(ii=inputs.begin(); ii!=inputs.end(); ) if(ii->problems()) { /*fprintf(stderr, "In killed\n");*/ problems = true; inputs.erase(ii); } else ++ii;
    
    vector<MixerOutput>::iterator oi;
    for(oi=outputs.begin(); oi!=outputs.end(); ) if(oi->problems()) { /*fprintf(stderr, "Out killed\n");*/ problems = true; outputs.erase(oi); } else ++oi;
    
    if(problems)
        waitpid(-1, NULL, WNOHANG);
    
    if(inputs.size())
    {
        unsigned least = 65536;
        for(ii=inputs.begin(); ii!=inputs.end(); ++ii)
        {
        	//if(ii->size())
	        //	fprintf(stderr, "Input %d has %u bytes to say\n", ii->getfd(), ii->size());
            if(ii->size() < least)
                least = ii->size();
        }
        /* If something to send */
        if(least)
        {
            stent Buf[least];
            memset(Buf, 0, sizeof Buf);
            
            for(ii=inputs.begin(); ii!=inputs.end(); ++ii)
            {
                unsigned a;
                const stent *source = ii->ptr();
                for(a=0; a<least; ++a)
                {
                    Buf[a].l += source[a].l;
                    Buf[a].r += source[a].r;
                }
                ii->Eat(least);
            }
            
            for(oi=outputs.begin(); oi!=outputs.end(); ++oi)
            {
            	//fprintf(stderr, "Sending %u bytes to output %d\n", least*4, oi->getfd());
                oi->Send(Buf, least);
            }
        }
    }
    else if(outputs.size() && !panic)
    {
        /* No inputs! Send hum. */
        char Buf[4096];
        int c = read(humfd, Buf, sizeof Buf);
        fprintf(stderr, "Humming %d bytes\n", c);
        if(c > 0)
        {
	        for(oi=outputs.begin(); oi!=outputs.end(); ++oi)
    	        if(!oi->panic())
        	        oi->Send((stent *)&Buf[0], c/4);
    	}
    }
}

#define VERSIONSTR \
    "mixer v"VERSION" - Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "This program is under GPL. wavegenerator-"VERSION".{rar,zip,tar.{gz,bz2}}\n" \
    "are available at the homepage of the author.\n"

class Handle : public arghandler
{
    bool Help;
private:
    string opt_h(const string &s)
    {
        Help = true;
        return s;
    }
    string opt_V(const string &)
    {
        printf(VERSIONSTR);
        exit(0);
    }
public:
    Handle(const char *defopts, int argc, const char *const *argv)
    : arghandler(defopts, argc, argv),
        Help(false)
    {
        add("-h", "--help",     "This text",    &(Handle::opt_h));
        add("-V", "--version",  "Version info", &(Handle::opt_V));
    }
    virtual void defarg(const string &)
    {
        //argerror(s, true);
    }
    void Usage()
    {
        fprintf(stderr,
            "Usage: %s [<options> ...]\n"
            "\n"
            "Provides a server for audio stream mixing.\n"
            "Meant to be a replacement for esd, with much more features\n"
            "and much less cpu usage. See mixwrite -h, mixmon -h\n"
            "\n"
            "Options for mixer:\n", a0.c_str());
        
        listoptions();
        
        exit(0);
    }
    virtual void parse()
    {
        arghandler::parse();
        if(Help)Usage();
    }
};

int main(int argc, const char *const *argv)
{
    Handle parameters("", argc, argv);
    parameters.parse();
    
    signal(SIGPIPE, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    
    Mixer mixer(MIXER_PORT);
    
    for(;;)
    {
        mixer.tick();
    }
    
    return 0;
}
